local debugUtil = require("debug.debugUtil")
local initChannel = love.thread.getChannel("myFfiInitChannel")
local syncChannel = love.thread.getChannel("myFfiSyncChannel")
-- GLOBALS: IS_MAIN_THREAD
--local needInit = initChannel:getCount() == 0
if IS_MAIN_THREAD then
  initChannel:push(true)
  syncChannel:push(true)
end




local function synchronized(baseFunction)
  assert(baseFunction, "no function to wrap")
  local capturedFunction = baseFunction
  return function(...)
    assert(initChannel:getCount()<=1, "init should be done once but done "..initChannel:getCount().." times")
    syncChannel:demand()
    local result = capturedFunction(...)
    syncChannel:push(true)
    return result
  end
end

local autoName = 1
local function debugged(baseFunction, name)
  assert(baseFunction)
  local capturedName
  if name then
    capturedName = tostring(name)
  else
    capturedName = "function_"..autoName
  end
  local capturedFunction = baseFunction
  return function(...)
    io.stdout:write("---CALL to "..capturedName.. " with args ")
    local args = {...}
    for i=1,#args do
      io.stdout:write(tostring(args[i])..";")
    end
    io.stdout:write("\n")
    io.stdout:flush()
    local result = capturedFunction(...)
    io.stdout:write("===END CALL to "..capturedName.. " with result "..tostring(result))
    io.stdout:write("\n")
    io.stdout:flush()
    return result
  end
end

local function wrap(baseFunction,name)
  return synchronized(debugged(baseFunction,name))
end
local function nowrap(baseFunction)
  return baseFunction
end




local function constructFfiWrapper()
  local ffi = require("ffi")
  ffi.cdef[[
    typedef struct {
      char *fpos;
      void *base;
      unsigned short handle;
      short flags;
      short unget;
      unsigned long alloc;
      unsigned short buffincrement;
    } FILE;

    FILE *fopen(const char *filename, const char *mode);
    int fprintf(FILE *stream, const char *format, ...);
    int fclose(FILE *stream);
    size_t fwrite(const void *ptr , size_t size, size_t nmemb, FILE *stream);
    size_t fseek(FILE *stream, size_t offset, int whence);
    size_t ftell(FILE *stream);
    size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream);

    void *malloc(size_t size);
    void free(void* memblock);
  ]]

  local SEEK_END = 2
  local SEEK_CUR = 1
  local SEEK_SET = 0
  local C = ffi.C

  local allocationTableChannel = love.thread.getChannel("allocationTableChannel")

  local function getN(ctype,ptr)
    local numberPtr = tonumber(ffi.cast("intptr_t", ptr))

    local allocationTable = allocationTableChannel:peek() or {}
    if not allocationTable[numberPtr] then
      debugUtil.trace(allocationTable,"allocation table")
      error("going to access something we didn't allocate! "..tostring(numberPtr))
    end
    local typeSize = ffi.sizeof(ctype)
    local n = allocationTable[numberPtr] / typeSize
    if n ~= math.floor(n) then error("allocated area size doesn't match type size!") end
    assert(n>0, "this mem has 0 entries!")
    return n
  end

  local function validateKey(key, ctype,ptr)
    if type(key)~= 'number' then error("key not number!") end
    if math.floor(key)~= key then error("key not int!") end
    if key < 0 or key >= getN(ctype, ptr) then error("access memory out of bounds!") end
    return key
  end

  local function castToSafePtr(ctype, ptr)
    --immediate check
    getN(ctype,ptr)

    local dataPtr = ffi.cast(ctype, ptr)
    --also need to be syncrhonized to avoid catching moments when allocation table is updated
    local safePtr = setmetatable({dataPtr=dataPtr, originalPtr=ptr, ctype=ctype}, {
       __index = wrap(function(t,k)
         return t.dataPtr[validateKey(k,  t.ctype, t.originalPtr)]
       end, "accessor"),
       __newindex = wrap(function(t,k,v)
         t.dataPtr[validateKey(k,  t.ctype, t.originalPtr)] = v
       end, "mutator")
    })



    return safePtr

  end

  local function safe_malloc(size)
    if not size then error("malloc without size!") end
    if size ~= math.floor(size) then error("malloc size not int!") end
    if size <=1 then error("malloc size less than 1!") end
    local result = C.malloc(size)
    if result == nil then
      error("malloc failed allocation!")
    end

    local function c(t)
      local count = 0
      for k,v in pairs(t) do count = count + 1 end
      return count
    end
    local temp = allocationTableChannel:pop()

    --debugUtil.trace(temp, "allocationpop")
    local allocationTable = temp or {}
    local oc = c(allocationTable)
    print("allocated size="..oc)
    allocationTable[tonumber(ffi.cast("intptr_t", result))] = size
    allocationTableChannel:push(allocationTable)
    assert(c(allocationTableChannel:peek())==oc+1)
    return result
  end

  local function safe_free(ptr)
    if type(ptr)=="table" then
      ptr = rawget(ptr, "originalPtr")
    end

    if ptr == nil then
      error("freeing nil!")
    end

    local numberPtr =  tonumber(ffi.cast("intptr_t", ptr))
    local allocationTable = allocationTableChannel:pop() or {}
    if not allocationTable[numberPtr] then error("freeing something we didn't allocate!") end
    allocationTable[numberPtr] = nil
    allocationTableChannel:push(allocationTable)
  end



  local myffi = {}
  myffi.cdef = nowrap(ffi.cdef, "cdef")
  myffi.cast = synchronized(ffi.cast, "cast")
  myffi.new = nowrap(ffi.new, "new ")
  myffi.metatype = nowrap(ffi.metatype, "metatype")
  myffi.string = nowrap(ffi.string, "string")
  myffi.copy = nowrap(ffi.copy, "copy")
  myffi.sizeof = nowrap(ffi.sizeof, "sizeof")
  myffi.C = {
    fopen = nowrap(C.fopen, "fopen"),
    fprintf = nowrap(C.fprintf, "fprintf"),
    fclose = nowrap(C.fclose, "fclose"),
    fwrite = nowrap(C.fwrite, "fwrite"),
    fseek = nowrap(C.fseek, "fseek"),
    ftell = nowrap(C.ftell, "ftell"),
    fread = nowrap(C.fread, "fread"),
    -- malloc = nowrap(safe_malloc, "malloc"),
    -- free = nowrap(safe_free, "free")
    malloc = C.malloc,
    free = C.free
  }

--  myffi.castToSafePtr = wrap(castToSafePtr, "castToSafePtr")

  myffi.SEEK_END = SEEK_END
  myffi.SEEK_CUR = SEEK_CUR
  myffi.SEEK_SET = SEEK_SET

  return myffi
end

local wrappedConstructFfiWrapper = nowrap(constructFfiWrapper, "constructFfiWrapper")
return wrappedConstructFfiWrapper()
